/* This file is part of the ili2fme project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.interlis.ili2fme;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import COM.safe.fme.pluginbuilder.IFMEWriter;
import COM.safe.fmeobjects.IFMEFeature;
import COM.safe.fmeobjects.IFMEFeatureVector;
import COM.safe.fme.pluginbuilder.IFMEMappingFile;
import COM.safe.fmeobjects.IFMESession;
import COM.safe.fmeobjects.FMEException;
import COM.safe.fmeobjects.IFMELogFile;

import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.ili2c.metamodel.*;

import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom.*;

/** INTERLIS implementation of an FME-Writer.
 * @author ce
 * @version $Revision: 1.0 $ $Date: 17.05.2005 $
 */
public class Ili2Writer implements IFMEWriter {
	private IFMEMappingFile mappingFile=null;
	private IFMESession session=null;
	private String writerKeyword=null;
	private String xtfFile=null;
	private IomFile iomFile=null;
	private ch.interlis.ili2c.metamodel.TransferDescription iliTd=null;
	private HashSet basketv=new HashSet(); // set<String basketId>
	private HashMap tag2class=null;
	public Ili2Writer(IFMESession session1,IFMEMappingFile mappingFile1,String keyword,IFMELogFile log){
		mappingFile=mappingFile1;
		writerKeyword=keyword;
		session=session1;
		setupLogging(log);
	}
	// Open up the writer and pass it the destination directory to which files will
	// be written (1st parameter) and a string array which may contain additional
	// dataset parameters (2nd parameter).
	public void open(ArrayList args) throws Exception {
		if(args.size()==0){
			throw new IllegalArgumentException("args.size()==0");
		}
		String models=null;
		String modeldir=null;
		for(int i=0;i<args.size();i++){
			String arg=(String)args.get(i);
			EhiLogger.traceState("arg["+Integer.toString(i)+"] "+arg);
			if(arg.equals(Ili2fme.MODELS)){
				i++;
				models=(String)args.get(i);
			}else if(arg.equals(Ili2fme.MODEL_DIR)){
				i++;
				modeldir=(String)args.get(i);
			}else if(arg.startsWith("+")){
				i++; // skip this and next(!) argument
			}else{
				// skip this argument
			}
		}
		Iterator elei=mappingFile.elements();
		while(elei.hasNext()){
			Object eleo=elei.next();
			EhiLogger.traceState("element "+eleo.getClass()+","+eleo);	
			if(eleo instanceof ArrayList){
				ArrayList ele=(ArrayList)eleo;
				String val=(String)ele.get(0);
				if(val.equals(writerKeyword+"_"+Ili2fme.MODELS)){
					models=(String)ele.get(1);	
				}else if(val.equals(writerKeyword+"_"+Ili2fme.MODEL_DIR)){
					modeldir=(String)ele.get(1);	
				}
			}
		}
		if(models==null){
			throw new IllegalArgumentException("model name not specified; set FME-Parameter Models");
		}
		if(modeldir==null){
			modeldir=new java.io.File(session.fmeHome(),"ilimodels").getAbsolutePath();
			EhiLogger.traceState("modeldir <"+modeldir+">");
		}
		
		xtfFile=(String)args.get(0);
		ArrayList iliModelv=null;
		if(models.equals("XTF")){
			throw new IllegalArgumentException("model name not specified; set FME-Parameter Models");
		}else{
			// parse string
			iliModelv=new ArrayList(java.util.Arrays.asList(models.split(";")));
		}
		// compile models
		String ilxfile=null;
		ilxfile=java.io.File.createTempFile("ili2fme","ilx").getAbsolutePath();
		iliTd=Ili2fme.compileIliModels(iliModelv,modeldir,ilxfile);
		if(iliTd==null){
			// compiler failed
			throw new IllegalArgumentException("compiler failed");
		}
		tag2class=ch.interlis.ili2c.generator.XSDGenerator.getTagMap(iliTd);
		// ready to write data
		setupIom();
		iomFile=ch.interlis.iom.Iom.openFile(xtfFile,iom_javaConstants.IOM_CREATE | iom_javaConstants.IOM_DONTREAD,null);
		// do writing
		IomFile iomIlx=ch.interlis.iom.Iom.openFile(ilxfile,0,null);
		iomFile.setmodel(iomIlx.iteratorbasket().nextbasket());

	}
	// Terminate the writer mid-stream.  Any special actions to shut
	// down a writer not finished writing data should be taken in this
	// method.  For most writers, nothing will be done.
	public void abort() throws Exception {
		iomFile=null;
		cleanup();
	}
	// Close the writer after it has completed writing out all the features.
	public void close() throws Exception {
		int saved=0;
		if(iomFile!=null){
			// check that each basket has a type
			Iterator basketi=basketv.iterator();
			while(basketi.hasNext()){
				String basketId=(String)basketi.next();
				IomBasket iomBasket=iomFile.getbasket(basketId);
				String basketType=iomBasket.getbaskettag();
				// has basket a type?
				if(basketType==null || basketType.length()==0){
					// no type
					// fix it
					IomObject obj=iomBasket.iteratorobject().nextobject();
					String objType=obj.getobjecttag();
					basketType=objType.substring(0,objType.lastIndexOf('.'));
					iomBasket.setbaskettag(basketType);
					EhiLogger.logAdaption("missing topic name of basket with BID="+basketId+"; fixed to "+basketType);
				}
			}
			saved=iomFile.save();
			iomFile.close();
			iomFile=null;
		}
		if(saved!=0 || listener.hasSeenErrors()){
			throw new Exception("INTERLIS 2 writer failed");
		}
		cleanup();
	}
	public int id() {
		// All Java writers return 0.
		return 0;
	}
	// Write the next feature to a file.
	public void write(IFMEFeature obj) 
	throws Exception 
	{
		try{
			// setup iom
			if(obj.getFeatureType().equals(Ili2fme.XTF_BASKETS)){
				//EhiLogger.debug("topic <"+obj.getStringAttribute(Ili2fme.XTF_TOPIC)+">, id <"+obj.getStringAttribute(Ili2fme.XTF_ID)+">");
				mapBasket(obj);
			}else{
				//EhiLogger.debug("class <"+obj.getStringAttribute(Ili2fme.XTF_CLASS)+">, id <"+obj.getStringAttribute(Ili2fme.XTF_ID)+">");
				mapFeature(obj,null,null,null);
			}
		}catch(Exception ex){
			EhiLogger.logError(ex);
			throw ex;
		}
	}
	private void mapBasket(IFMEFeature obj)
	throws Exception 
	{
		String topic=obj.getStringAttribute(Ili2fme.XTF_TOPIC);
		String basketId=obj.getStringAttribute(Ili2fme.XTF_ID);
		IomBasket iomBasket=getBasket(basketId);
		iomBasket.setbaskettag(topic);
	}
	private void mapFeature(IFMEFeature obj,String fmeListAttrPrefix,IomObject iliStructParent,String iliStructAttrName)
	throws Exception 
	{
		boolean isStructEle=fmeListAttrPrefix!=null;
		String attrPrefix="";
		String fmeRecInfo=obj.getFeatureType();
		if(isStructEle){
			fmeRecInfo = fmeRecInfo+" "+fmeListAttrPrefix;
			attrPrefix=fmeListAttrPrefix+".";
		}
		if(!obj.attributeExists(attrPrefix+Ili2fme.XTF_CLASS)){
			String err=fmeRecInfo+": missing mandatory attribute "+attrPrefix+Ili2fme.XTF_CLASS;
			EhiLogger.logError(err);
			throw new Exception(err);
		}
		String tag=obj.getStringAttribute(attrPrefix+Ili2fme.XTF_CLASS);
		if(tag==null || tag.length()==0){
			String err=fmeRecInfo+": missing mandatory attribute "+attrPrefix+Ili2fme.XTF_CLASS;
			EhiLogger.logError(err);
			throw new Exception(err);
		}
		IomObject iomObj=null;
		if(!isStructEle){
			if(!obj.attributeExists(Ili2fme.XTF_BASKET)){
				String err=fmeRecInfo+": missing mandatory attribute "+Ili2fme.XTF_BASKET;
				EhiLogger.logError(err);
				throw new Exception(err);
			}
			if(!obj.attributeExists(Ili2fme.XTF_ID)){
				String err=fmeRecInfo+": missing mandatory attribute "+Ili2fme.XTF_ID;
				EhiLogger.logError(err);
				throw new Exception(err);
			}
			String basketId=obj.getStringAttribute(Ili2fme.XTF_BASKET);
			if(basketId==null || basketId.length()==0){
				String err=fmeRecInfo+": missing mandatory attribute "+Ili2fme.XTF_BASKET;
				EhiLogger.logError(err);
				throw new Exception(err);
			}
			IomBasket iomBasket=getBasket(basketId);
			String tid=obj.getStringAttribute(Ili2fme.XTF_ID);
			if(tid==null || tid.length()==0){
				String err=fmeRecInfo+": missing mandatory attribute "+Ili2fme.XTF_ID;
				EhiLogger.logError(err);
				throw new Exception(err);
			}
			iomObj=iomBasket.newobject(tag,tid);	
		}else{
			iomObj=iliStructParent.addattrobj(iliStructAttrName,tag);	
		}
		Viewable aclass=(Viewable)tag2class.get(tag);
		if(aclass==null){
			String err=fmeRecInfo+": unknown interlis class <"+tag+">";
			EhiLogger.logError(err);
			throw new Exception(err);
			
		}
		Iterator iter = aclass.getAttributesAndRoles2();
		String geomattr=null;
		if(!isStructEle && obj.attributeExists(Ili2fme.XTF_GEOMATTR)){
			geomattr=obj.getStringAttribute(Ili2fme.XTF_GEOMATTR);
		}
		while (iter.hasNext()) {
			Viewable.TransferElement prop = (Viewable.TransferElement)iter.next();
			if (prop.obj instanceof AttributeDef) {
				AttributeDef attr = (AttributeDef) prop.obj;
				Type type = attr.getDomainResolvingAliases();
				 String attrName=attr.getName();
				if (type instanceof CompositionType){
					int elei=0;
					String prefix=attrPrefix+attrName+"{"+Integer.toString(elei)+"}";
					while(obj.attributeExists(prefix+"."+Ili2fme.XTF_CLASS)){
						// struct element
						//EhiLogger.debug("struct "+prefix);
						mapFeature(obj,prefix,iomObj,attrName);						
						elei++;
						prefix=attrPrefix+attrName+"{"+Integer.toString(elei)+"}";
					}
				}else if (type instanceof PolylineType){
					if(geomattr!=null && attrName.equals(geomattr)){
						boolean is3D=3==((CoordType)((PolylineType)type).getControlPointDomain().getType()).getDimensions().length;
						IomObject polylineValue=iomObj.addattrobj(attrName,"POLYLINE");
						// unclipped polyline, add one sequence
						IomObject sequence=polylineValue.addattrobj("sequence","SEGMENTS");
						int segc=obj.numCoords();
						for(int segi=0;segi<segc;segi++){
							// add control point
							double valuev[]=obj.get3DCoordinate(segi);
							IomObject coordValue=sequence.addattrobj("segment","COORD");
							coordValue.setattrvalue("C1",Double.toString(valuev[0]));
							coordValue.setattrvalue("C2",Double.toString(valuev[1]));
							if(is3D){
								coordValue.setattrvalue("C3",Double.toString(valuev[2]));
							}
							//coordValue.setattrvalue("A1",Double.toString(ordv[i]));
							//coordValue.setattrvalue("A2",Double.toString(ordv[i+1]));
							//if(is3D){
							//	// no A3 in XTF!
							//}
						}
					}
				 }else if(type instanceof SurfaceOrAreaType){
					if(geomattr!=null && attrName.equals(geomattr)){
						boolean is3D=3==((CoordType)((SurfaceOrAreaType)type).getControlPointDomain().getType()).getDimensions().length;
						//fmeLog.logFeature(obj,IFMELogFile.FME_INFORM,-1);
						IFMEFeatureVector bndries=session.createFeatureVector();
						obj.getDonutParts(bndries);
						IomObject multisurface=iomObj.addattrobj(attrName,"MULTISURFACE");
						IomObject surface=multisurface.addattrobj("surface","SURFACE");
						int bndc=bndries.entries();
						//EhiLogger.debug("number of boundaries "+Integer.toString(bndc));
						for(int bndi=0;bndi<bndc;bndi++){
							IFMEFeature line=bndries.getAt(bndi);
							IomObject boundary=surface.addattrobj("boundary","BOUNDARY");
							IomObject polylineValue=boundary.addattrobj("polyline","POLYLINE");
							IomObject sequence=polylineValue.addattrobj("sequence","SEGMENTS");
							int segc=line.numCoords();
							for(int segi=0;segi<segc;segi++){
								// add control point
								double valuev[]=line.get3DCoordinate(segi);
								IomObject coordValue=sequence.addattrobj("segment","COORD");
								coordValue.setattrvalue("C1",Double.toString(valuev[0]));
								coordValue.setattrvalue("C2",Double.toString(valuev[1]));
								if(is3D){
									coordValue.setattrvalue("C3",Double.toString(valuev[2]));
								}
								//coordValue.setattrvalue("A1",Double.toString(ordv[i]));
								//coordValue.setattrvalue("A2",Double.toString(ordv[i+1]));
								//if(is3D){
								//	// no A3 in XTF!
								//}
							}
						}
					}
				 }else if(type instanceof CoordType){
					if(geomattr!=null && attrName.equals(geomattr)){
						double valuev[]=obj.get3DCoordinate(0);
						IomObject coord=iomObj.addattrobj(attrName,"COORD");
						coord.setattrvalue("C1",Double.toString(valuev[0]));
						coord.setattrvalue("C2",Double.toString(valuev[1]));
						if(((CoordType)type).getDimensions().length==3){
							coord.setattrvalue("C3",Double.toString(valuev[2]));
						}
						
					}
				}else{
					if(obj.attributeExists(attrPrefix+attrName)){
						String value=obj.getStringAttribute(attrPrefix+attrName);
						if(value!=null){
							iomObj.setattrvalue(attrName, value);
						}
					}else{
					}
				}
			}
			if(prop.obj instanceof RoleDef){
				RoleDef role = (RoleDef) prop.obj;
				String roleName=role.getName();
				// a role of an embedded association?
				if(prop.embedded){
					AssociationDef roleOwner = (AssociationDef) role.getContainer();
					if(roleOwner.getDerivedFrom()==null){
						// not just a link?
						if (roleOwner.getAttributes().hasNext()
							|| roleOwner.getLightweightAssociations().iterator().hasNext()) {
							 // TODO add association attributes
						}
						if(obj.attributeExists(roleName)){
							String refoid=obj.getStringAttribute(roleName);
							IomObject structvalue=iomObj.addattrobj(roleName,"REF");
							 structvalue.setobjectrefoid(refoid);
							 if(role.isOrdered()){
								long orderPos=obj.getIntAttribute(roleName+"."+Ili2fme.ORDERPOS);
								structvalue.setobjectreforderpos(orderPos);
							 }
						}
					}
				}else{
					String refoid=obj.getStringAttribute(roleName);
					IomObject structvalue=iomObj.addattrobj(roleName,"REF");
					 structvalue.setobjectrefoid(refoid);
					 if(role.isOrdered()){
						long orderPos=obj.getIntAttribute(roleName+"."+Ili2fme.ORDERPOS);
						structvalue.setobjectreforderpos(orderPos);
					 }
				}
			}
		}
			
	}
	/** get or create basket with given basket id
	 */
	private IomBasket getBasket(String basketId)
	{
		IomBasket iomBasket=null;
		// basket not yet seen?
		if(!basketv.contains(basketId)){
			// create new basket
			iomBasket=iomFile.newbasket();
			iomBasket.setbasketoid(basketId);
			// add to list of seen baskets
			basketv.add(basketId);
		}else{
			// get existing basket
			iomBasket=iomFile.getbasket(basketId);
		}
		return iomBasket;
	}
	public boolean multiFileWriter() {
		// don't create a new object for every classdef
		return false;
	}
	public void startTransaction() throws Exception {
		//For formats or systems which do not have the notion of a transaction then nothing needs to be done.
	}
	public void commitTransaction() throws Exception {
		//For formats or systems which do not have the notion of a transaction then nothing needs to be done.
	}
	public void rollbackTransaction() throws Exception {
		//For formats or systems which do not have the notion of a transaction then nothing needs to be done.
	}
	private FmeLogListener listener=null;
	//private Log2File listener2=null;
	private void setupLogging(IFMELogFile logFile)
	{
		if(listener==null){
			listener=new FmeLogListener(logFile);
			EhiLogger.getInstance().addListener(listener);
			//listener2=new Log2File("c:/tmp/ili2fme/ce.log");
			//EhiLogger.getInstance().addListener(listener2);
			EhiLogger.getInstance().addListener(ch.ehi.basics.logging.StdListener.getInstance());
			EhiLogger.getInstance().removeListener(ch.ehi.basics.logging.StdListener.getInstance());
			//EhiLogger.getInstance().setTraceFiler(false);
		}
		
	}
	private boolean doSetupIom=true;
	private void setupIom(){
		// setup iom
		if(doSetupIom){
			EhiLogger.traceState("setup IOM");
			ch.interlis.iom.Iom.setLibraryHint(Ili2fme.ILI2FME_JAR);
			ch.interlis.iom.Iom.init();
			IomErrorListener iomerrs=new IomErrorListener();
			ch.interlis.iom.Iom.addErrListener(iomerrs);
			doSetupIom=false;
		}
	}
	private void cleanup(){
		// did setup iom?
		if(!doSetupIom){
			ch.interlis.iom.Iom.end();
			doSetupIom=true;
		}
		if(listener!=null){
			EhiLogger.getInstance().removeListener(listener);
			listener=null;
		}
		//if(listener2!=null){
		//	EhiLogger.getInstance().removeListener(listener2);
		//	listener2=null;
		//}
	}
}
