/* This file is part of the ili2fme project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.interlis.ili2fme;

import COM.safe.fme.pluginbuilder.IFMEMappingFile;
import COM.safe.fme.pluginbuilder.IFMEReader;
import COM.safe.fme.pluginbuilder.IFMEReaderCreator;
import COM.safe.fme.pluginbuilder.IFMEWriter;
import COM.safe.fme.pluginbuilder.IFMEWriterCreator;
import COM.safe.fmeobjects.IFMECoordSysManager;
import COM.safe.fmeobjects.IFMELogFile;
import COM.safe.fmeobjects.IFMESession;

import ch.ehi.basics.logging.EhiLogger;
import java.util.ArrayList;
import ch.interlis.ili2c.metamodel.TransferDescription;

/** Factory called by FME to create the reader/writer.
 * @author ce
 * @version $Revision: 1.0 $ $Date: 20.01.2005 $
 */
public class Ili2fme implements IFMEReaderCreator, IFMEWriterCreator {
	static public final String XTF_ID="xtf_id";
	static public final String XTF_BASKET="xtf_basket";
	static public final String XTF_BASKETS="XTF_BASKETS";
	static public final String XTF_TOPIC="xtf_topic";
	static public final String XTF_CLASS="xtf_class";
	static public final String XTF_GEOMATTR="xtf_geomattr";
	static public final String XTF_GEOMTYPE="xtf_geomtype";
	static public final String ORDERPOS="orderPos";
	static public final String MODELS="MODELS";
	static public final String MODEL_DIR="MODEL_DIR";
	static public final String ID_TYPE="xtf_char(200)";
	static public final String ILINAME_TYPE="xtf_char(255)";
	static public final String ILIQNAME_TYPE="xtf_char(757)"; // 3*255+2 "Model.Topic.Class"
	static public final String ILI2FME_JAR="ili2fme.jar";
	/** compiles a set of ili models.
	 */
	static public TransferDescription compileIliModels(ArrayList modelv,String modeldir,String ilxFile){
		ch.interlis.ili2c.config.Configuration config=ch.interlis.ili2c.ModelScan.getConfig(modeldir,modelv);
		if(config==null){
			return null;
		}
		if(ilxFile!=null){
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.IOM);
			config.setOutputFile(ilxFile);
		}else{
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.NOOUTPUT);
		}
		TransferDescription ret=ch.interlis.ili2c.Main.runCompiler(config);
		return ret;
	}
	/** compiles a set of ili files.
	 */
	static public TransferDescription compileIliFiles(ArrayList filev,String modeldir,String ilxFile){
		ArrayList modeldirv=new ArrayList();
		modeldirv.add(modeldir);
		ch.interlis.ili2c.config.Configuration config=ch.interlis.ili2c.ModelScan.getConfigWithFiles(modeldirv,filev);
		if(config==null){
			return null;
		}
		if(ilxFile!=null){
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.IOM);
			config.setOutputFile(ilxFile);
		}else{
			config.setOutputKind(ch.interlis.ili2c.config.GenerateOutputKind.NOOUTPUT);
		}
		TransferDescription ret=ch.interlis.ili2c.Main.runCompiler(config);
		return ret;
	}

	public IFMEReader createReader(IFMEMappingFile mappingFile,
			IFMELogFile logFile,
			IFMECoordSysManager coordSysMan,
			IFMESession session,
			String readerTypeName,
			String readerKeyword)

		throws Exception 
	{
		//logFile.logMessageString("createReader()",IFMELogFile.FME_INFORM);
		Ili2Reader ret=null;
		try{	
			ret=new Ili2Reader(session,mappingFile,readerKeyword,logFile);
		}catch(Exception ex){
			logFile.logMessageString("createReader() "+ex.getMessage(),IFMELogFile.FME_ERROR);
			throw ex;
		}
		return ret;
	}

	public IFMEWriter createWriter(
				IFMEMappingFile mappingFile,
				IFMELogFile logFile,
				IFMECoordSysManager coordSysMan,
				IFMESession session,
				String readerTypeName,
				String readerKeyword)
		throws Exception 
	{
		Ili2Writer ret=null;
		try{
			ret=new Ili2Writer(session,mappingFile,readerKeyword,logFile);
		}catch(Exception ex){
			logFile.logMessageString("createWriter()"+ex.getMessage(),IFMELogFile.FME_ERROR);
			throw ex;
		}
		return ret;
	}
	private static String version=null;
	public static String getVersion() {
		  if(version==null){
		java.util.ResourceBundle resVersion = java.util.ResourceBundle.getBundle("ch/interlis/ili2fme/Version");
			// Major version numbers identify significant functional changes.
			// Minor version numbers identify smaller extensions to the functionality.
			// Micro versions are even finer grained versions.
			StringBuffer ret=new StringBuffer(20);
		ret.append(resVersion.getString("versionMajor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMinor"));
			ret.append('.');
		ret.append(resVersion.getString("versionMicro"));
			ret.append('-');
		ret.append(resVersion.getString("versionDate"));
			version=ret.toString();
		  }
		  return version;
	}
}
