/*============================================================================

   Name     : center.java

   System   : FME Developer Kit

   Language : Java

   Purpose  : Implements the IFMEFunction interface.

   Author               Date            Changes made
   ------------------   ------------    -------------------------------
   Your Name            Xxx XX, 200X    Original definition

                  Copyright (c) 1999 - 2000, Your Company Goes Here
                              All Rights Reserved

    The entire risk as to the results and performance of the software,
    supporting text and other information contained in this file
    (collectively called the "Software") is with the user.  Although
    Safe Software Incorporated has used considerable efforts in preparing
    the Software, Safe Software Incorporated does not warrant the
    accuracy or completeness of the Software. In no event will Safe Software
    Incorporated be liable for damages, including loss of profits or
    consequential damages, arising out of the use of the Software.

    NOTE: Replace the above header with whatever is appropriate for your
    company.

============================================================================*/

package com.mycompany.myfunction;

import COM.safe.fme.pluginbuilder.*;
import COM.safe.fmeobjects.*;
import java.util.*;

public class center implements IFMEFunction
{
   // Private instance variables.
   private IFMEMappingFile mappingFile_ = null;
   private IFMELogFile logFile_ = null;
   private IFMECoordSysManager coordSysMan_ = null;
   private IFMESession session_ = null;

   //===========================================================================
   // Constructor
   public center() throws Exception
   {
      // do nothing
   }

   //===========================================================================
   // initialize
   public void initialize(IFMEMappingFile mappingFile,
                          IFMELogFile logFile,
                          IFMECoordSysManager coordSysMan,
                          IFMESession session)
                          throws Exception
   {
      mappingFile_ = mappingFile;
      logFile_     = logFile;
      coordSysMan_ = coordSysMan;
      session_     = session;
   }

   //===========================================================================
   // close
   public void close() throws Exception
   {
      // do nothing
   }

   //===========================================================================
   // abort
   public void abort() throws Exception
   {
      // do nothing
   }

   //===========================================================================
   // execute
   public String execute(IFMEFeature feature, ArrayList parameterList) throws Exception
   {
      // All we will do is average all the coordinates and then set the
      // various attributes to the right values

      double xSum = 0;
      double ySum = 0;
      double zSum = 0;

      int coordNum = 0;

      double startX = 0;
      double startY = 0;
      double startZ = 0;
      boolean getStart = true;

      String ret = "";

      boolean doingPolygon = (boolean)(
                    (feature.getGeometryType() == IFMEFeature.FME_GEOM_POLYGON) ||
                    (feature.getGeometryType() == IFMEFeature.FME_GEOM_DONUT));

      int originalNumCoords = feature.numCoords();
      int denominator = feature.numCoords();


      for (coordNum = 0; coordNum < originalNumCoords; coordNum++)
      {
         double xValue = 0;
         double yValue = 0;
         double zValue = 0;

         if (feature.getDimension() == 3)
         {
            xValue = feature.getXCoordinate(coordNum);
            yValue = feature.getYCoordinate(coordNum);
            zValue = feature.getZCoordinate(coordNum);
         }
         else
         {
            xValue = feature.getXCoordinate(coordNum);
            yValue = feature.getYCoordinate(coordNum);
         }

         if ((!doingPolygon) ||
            (getStart)      ||
            (startX != xValue) ||
            (startY != yValue))
         {
            xSum += xValue;
            ySum += yValue;
            zSum += zValue;

            if (getStart)
            {
               startX = xValue;
               startY = yValue;
               startZ = zValue;
               getStart = false;
            }
         }
         else
         {
            // the only way in here is if we hit the end of a ring and we
            // were doing a polygon.  Need to get the start point of the
            // next ring on the next iteration through
            getStart = true;
            denominator--;
         }
      }

      // if the denominator is zero, we throw an exception.
      if (denominator != 0)
      {
         feature.setDoubleAttribute((String)(parameterList.get(0)), xSum / denominator);
         feature.setDoubleAttribute((String)(parameterList.get(1)), ySum / denominator);
         feature.setDoubleAttribute((String)(parameterList.get(2)), zSum / denominator);
      }
      else
      {
         throw new Exception("Attempted to divide by zero.");
      }

      return ret;
   }

   //===========================================================================
   // inverse
   public String inverse(IFMEFeature feature, ArrayList parameterList) throws Exception
   {
      // The inverse method does the same as the execute method
      return execute(feature, parameterList);
   }
}